#ifndef __FT2232JTAG_H__
#define __FT2232JTAG_H__

#include "cable.h"
#include "long_bits.h"

#ifdef WIN32

#define USE_FT2232JTAG

#include "FTCJTAG/FTCJTAG.h"

struct pin_state_t {
  BOOL &in_out;
  BOOL &low_high;
  pin_state_t(BOOL &_in_out, BOOL &_low_high) : in_out(_in_out), low_high(_low_high) {}
  ~pin_state_t(){}
  static const BOOL dir_out = true;
  static const BOOL dir_in = false;
  static const BOOL level_high = true;
  static const BOOL level_low = false;
};

class FT2232JTAG : public cable {
  protected:
    FTC_HANDLE ftHandle;
    
    DWORD dwDeviceNameIndex;
    DWORD dwClockFrequencyHz;
    DWORD dwClockDivisor;
    DWORD dwTapNext;
    
    FTC_INPUT_OUTPUT_PINS LowInputOutputPinsData;
    FTC_INPUT_OUTPUT_PINS HighInputOutputPinsData;
    FTC_LOW_HIGH_PINS LowPinsInputData;
    FTC_LOW_HIGH_PINS HighPinsInputData;
  
  protected:
    virtual const pin_state_t *jtag_oe_n () const = 0;
    virtual const pin_state_t *vref_n_in () const = 0;
    virtual const pin_state_t *trst_n_oe_n () const = 0;
    virtual const pin_state_t *trst_n_out () const = 0;
    virtual const pin_state_t *srst_n_oe_n () const = 0;
    virtual const pin_state_t *srst_n_out () const = 0;
    virtual const pin_state_t *srst_n_in () const = 0;
  
  public:
    FT2232JTAG(const char* param) 
        : cable(), 
          dwDeviceNameIndex(0), dwClockDivisor(0) {
      
      LowInputOutputPinsData.bPin1InputOutputState = false; // = input
      LowInputOutputPinsData.bPin2InputOutputState = false;
      LowInputOutputPinsData.bPin3InputOutputState = false;
      LowInputOutputPinsData.bPin4InputOutputState = false;
      LowInputOutputPinsData.bPin1LowHighState = false;     // = low
      LowInputOutputPinsData.bPin2LowHighState = false;
      LowInputOutputPinsData.bPin3LowHighState = false;
      LowInputOutputPinsData.bPin4LowHighState = false;

      HighInputOutputPinsData.bPin1InputOutputState = false;
      HighInputOutputPinsData.bPin2InputOutputState = false;
      HighInputOutputPinsData.bPin3InputOutputState = false;
      HighInputOutputPinsData.bPin4InputOutputState = false;
      HighInputOutputPinsData.bPin1LowHighState = false;
      HighInputOutputPinsData.bPin2LowHighState = false;
      HighInputOutputPinsData.bPin3LowHighState = false;
      HighInputOutputPinsData.bPin4LowHighState = false;
      
      if(param){ // [[nakatomo]]
        char *dummy;
        dwClockDivisor = strtol(param, &dummy, 10);
        tracef(2, "FT2232: dwClockDivisor = %d; clock = %f Hz\n", 
            dwClockDivisor, 
            (double)(12e6) / ((1 + dwClockDivisor) * 2));
      }
    }
    
    virtual ~FT2232JTAG() {
      close();
    }

  public:
    cable::res_t open(){
      FTC_STATUS Status;
      
      {
        char szDllVersion[10];
        Status = JTAG_GetDllVersion(szDllVersion, 10);
        if(Status != FTC_SUCCESS){return cable::res_fail;}
      }
    
      {
        DWORD dwNumDevices = 0;
        Status = JTAG_GetNumDevices(&dwNumDevices);
        if((Status != FTC_SUCCESS) || (dwNumDevices <= 0)){return cable::res_fail;}
      }
      
      {
        char szDeviceName[100];
        DWORD dwLocationID = 0;
        
        Status = JTAG_GetDeviceNameLocID(
            dwDeviceNameIndex, szDeviceName, sizeof(szDeviceName), &dwLocationID);
        if(Status != FTC_SUCCESS){return cable::res_fail;}
     
        Status = JTAG_OpenEx(szDeviceName, dwLocationID, &ftHandle);
        if(Status != FTC_SUCCESS){return cable::res_fail;}
      }
      
      Status = JTAG_InitDevice(ftHandle, dwClockDivisor);
      if(Status != FTC_SUCCESS){return cable::res_fail;}
      dwTapNext = RUN_TEST_IDLE_STATE;
      
      Status = JTAG_GetClock(dwClockDivisor, &dwClockFrequencyHz);
      if(Status != FTC_SUCCESS){return cable::res_fail;}
      
      //Status = JTAG_SetLoopback(ftHandle, true);
      //if (Status != FTC_SUCCESS){return cable::res_fail;}
      
      // GPIO setting
      {
        const pin_state_t *state;
        if(state = jtag_oe_n()){
          state->in_out = pin_state_t::dir_out;
          state->low_high = pin_state_t::level_low;
        }
        if(state = vref_n_in()){
          state->in_out = pin_state_t::dir_in;
        }
        if(state = trst_n_oe_n()){
          state->in_out = pin_state_t::dir_out;
          state->low_high = pin_state_t::level_low;
        }
        if(state = trst_n_out()){
          state->in_out = pin_state_t::dir_out;
          state->low_high = pin_state_t::level_low;
        }
        if(state = srst_n_oe_n()){
          state->in_out = pin_state_t::dir_out;
          state->low_high = pin_state_t::level_low;
        }
        if(state = srst_n_out()){
          state->in_out = pin_state_t::dir_out;
          state->low_high = pin_state_t::level_low;
        }
        if(state = srst_n_in()){
          state->in_out = pin_state_t::dir_in;
        }
      }
    
      Status = JTAG_SetGPIOs(ftHandle, true, &LowInputOutputPinsData, true, &HighInputOutputPinsData);
      if(Status != FTC_SUCCESS){return cable::res_fail;}
      
      Status = JTAG_GetGPIOs(ftHandle, true, &LowPinsInputData, true, &HighPinsInputData);
      if(Status != FTC_SUCCESS){return cable::res_fail;}
      
      return cable::res_success;
    }
    cable::res_t close(){
      if(!ftHandle){  // [[nakatomo]]
        return cable::res_success;
      }
      FTC_STATUS Status = JTAG_Close(ftHandle);
      return (Status == FTC_SUCCESS) ? cable::res_success : cable::res_fail;
    }

    int get_desc(char* desc, int len){
      FTC_STATUS Status;

      char szDeviceName[100];
      DWORD dwLocationID = 0;
        
      Status = JTAG_GetDeviceNameLocID(
          dwDeviceNameIndex, szDeviceName, sizeof(szDeviceName), &dwLocationID);
      
      return snprintf(desc, len, "%s", 
          (Status == FTC_SUCCESS) ? szDeviceName : "FT2232 based cable");
    }

    //virtual int set_option(const char* name, const char* value);
    //virtual int get_option(const char* name, char* value, int len);
    
    cable::res_t set_tms(int b){
      return cable::res_success;
    }
    cable::res_t set_tck(int b){
      return cable::res_success;
    }
    cable::res_t pulse_tck(int count){
      FTC_STATUS Status = JTAG_GenerateClockPulses(ftHandle, count);
      return (Status == FTC_SUCCESS) ? cable::res_success : cable::res_fail;
    }
    
    cable::res_t state_transit(int state_from, int state_to){
      switch(state_to){
        case TAPSTATE_RESET:
          dwTapNext = TEST_LOGIC_STATE;
          break;
        case TAPSTATE_EXIT1DR:
        case TAPSTATE_EXIT2DR:
        case TAPSTATE_UPDATEDR:
        case TAPSTATE_EXIT1IR:
        case TAPSTATE_EXIT2IR:
        case TAPSTATE_UPDATEIR:
        case TAPSTATE_IDLE:
          dwTapNext = RUN_TEST_IDLE_STATE;
          break;
        case TAPSTATE_PAUSEDR:
          dwTapNext = PAUSE_TEST_DATA_REGISTER_STATE;
          break;
        case TAPSTATE_PAUSEIR:
          dwTapNext = PAUSE_INSTRUCTION_REGISTER_STATE;
          break;
        case TAPSTATE_SELECTDR:
        case TAPSTATE_CAPTUREDR:
        case TAPSTATE_SHIFTDR:
          dwTapNext = SHIFT_TEST_DATA_REGISTER_STATE;
          break;
        case TAPSTATE_SELECTIR:
        case TAPSTATE_CAPTUREIR:
        
        case TAPSTATE_SHIFTIR:
          dwTapNext = SHIFT_INSTRUCTION_REGISTER_STATE;
          break;
      }
      return cable::res_success;
    }
    
    // [[nakatomo]]
    cable::res_t shift(
        int wnum_bits, const void* wbuff, 
        int rnum_bits, void* rbuff, bool exit_shift_state){
      BOOL bInstructionTestData;
      switch(dwTapNext){
        case PAUSE_TEST_DATA_REGISTER_STATE:
        case SHIFT_TEST_DATA_REGISTER_STATE:
          bInstructionTestData = FALSE;
          break;
        case PAUSE_INSTRUCTION_REGISTER_STATE:
        case SHIFT_INSTRUCTION_REGISTER_STATE:
          bInstructionTestData = TRUE;
          break;
        default:
          return cable::res_fail;
      }
      
      if(exit_shift_state){
        dwTapNext = RUN_TEST_IDLE_STATE;
      }else{
        switch(dwTapNext){
          case SHIFT_TEST_DATA_REGISTER_STATE:
            dwTapNext = PAUSE_TEST_DATA_REGISTER_STATE;
            break;
          case SHIFT_INSTRUCTION_REGISTER_STATE:
            dwTapNext = PAUSE_INSTRUCTION_REGISTER_STATE;
            break;
        }
      }
      
      tracef(3, "FT2232: shift(): inst=%d, wnum_bits=%d\n", bInstructionTestData, wnum_bits);
      tracef(3, "FT2232: wbuff: ");
      trace(3, LongBits<unsigned char>((unsigned char *)wbuff, wnum_bits));
      tracef(3, "\n");
      
      ReadDataByteBuffer ReadDataBuffer;
      DWORD dwNumBytesReturned = 0;
      FTC_STATUS Status;
      
      // without comparison
      if(rnum_bits == 0){
        // should be w_num_bits > 0 or r_num_bits > 0
        if(wnum_bits == 0){
          tracef(2, "FT2232: rnum_bits == 0 && wnum_bits == 0\n");
          return cable::res_fail;
        }
        Status = JTAG_Write(ftHandle, 
            bInstructionTestData, wnum_bits, 
            (PWriteDataByteBuffer)wbuff, (wnum_bits + 8 - 1) / 8, 
            dwTapNext);
      }else if(wnum_bits == 0){
        Status = JTAG_Read(ftHandle, 
            bInstructionTestData, rnum_bits, 
            &ReadDataBuffer, &dwNumBytesReturned, dwTapNext);
      }else{
        // should be w_num_bits == r_num_bits
        if(wnum_bits != rnum_bits){
          tracef(2, "FT2232: rnum_bits != 0 && wnum_bits(%d) != rnum_bits(%d)\n", wnum_bits, rnum_bits);
          return cable::res_fail; //rnum_bits = wnum_bits; // another solution?
        }
  
        Status = JTAG_WriteRead(ftHandle, 
            bInstructionTestData, wnum_bits, (PWriteDataByteBuffer)wbuff, (wnum_bits + 8 - 1) / 8, 
            &ReadDataBuffer, &dwNumBytesReturned, dwTapNext);
      }
      
      tracef(3, "FT2232: shift(): finished; code=%d, returned byte=%d\n", Status, dwNumBytesReturned);
      tracef(3, "FT2232: rbuff: ");
      trace(3, LongBits<unsigned char>((unsigned char *)ReadDataBuffer, dwNumBytesReturned * 8));
      tracef( 3, "\n" );
      
      if(Status != FTC_SUCCESS){
        return cable::res_fail;
      }
      
      if(rbuff){
        memcpy(rbuff, &ReadDataBuffer, dwNumBytesReturned);
      }
      
      return cable::res_success;
    }
    
    // dispatcher
    cable::res_t shift(int num_bits, const void* wbuff, void* rbuff, bool exit_shift_state){
      return shift(num_bits, wbuff, num_bits, rbuff, exit_shift_state);
    }
};

class AmontecJTAGKey : public FT2232JTAG {
  protected:
    pin_state_t _jtag_oe_n;
    pin_state_t _vref_n_in;
    pin_state_t _trst_n_oe_n;
    pin_state_t _trst_n_out;
    pin_state_t _srst_n_oe_n;
    pin_state_t _srst_n_out;
    pin_state_t _srst_n_in;
  public:
    AmontecJTAGKey(const char* param) 
        : FT2232JTAG(param),
          _jtag_oe_n(   // 20 => GPIOL0
              FT2232JTAG::LowInputOutputPinsData.bPin1InputOutputState,
              FT2232JTAG::LowInputOutputPinsData.bPin1LowHighState),
          _vref_n_in(   // 19 => GPIOL1
              FT2232JTAG::LowInputOutputPinsData.bPin2InputOutputState,
              FT2232JTAG::LowInputOutputPinsData.bPin2LowHighState),
          _trst_n_oe_n( // 12 => GPIOH2
              FT2232JTAG::HighInputOutputPinsData.bPin3InputOutputState,
              FT2232JTAG::HighInputOutputPinsData.bPin3LowHighState),
          _trst_n_out(  // 15 => GPIOH0
              FT2232JTAG::HighInputOutputPinsData.bPin1InputOutputState,
              FT2232JTAG::HighInputOutputPinsData.bPin1LowHighState),
          _srst_n_oe_n( // 11 => GPIOH3
              FT2232JTAG::HighInputOutputPinsData.bPin4InputOutputState,
              FT2232JTAG::HighInputOutputPinsData.bPin4LowHighState),
          _srst_n_out(  // 13 => GPIOH1
              FT2232JTAG::HighInputOutputPinsData.bPin2InputOutputState,
              FT2232JTAG::HighInputOutputPinsData.bPin2LowHighState),
          _srst_n_in(   // 17 => GPIOL2
              FT2232JTAG::LowInputOutputPinsData.bPin3InputOutputState,
              FT2232JTAG::LowInputOutputPinsData.bPin3LowHighState){
    
    
    }
    ~AmontecJTAGKey(){}
    
    const pin_state_t *jtag_oe_n()    const {return &_jtag_oe_n;}
    const pin_state_t *vref_n_in()    const {return &_vref_n_in;} 
    const pin_state_t *trst_n_oe_n()  const {return &_trst_n_oe_n;}
    const pin_state_t *trst_n_out()   const {return &_trst_n_out;}
    const pin_state_t *srst_n_oe_n()  const {return &_srst_n_oe_n;}
    const pin_state_t *srst_n_out()   const {return &_srst_n_out;}
    const pin_state_t *srst_n_in()    const {return &_srst_n_in;}
};

#endif

#endif /* __FT2232JTAG_H__ */
